/*
 * Copyright (C) 2014 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of version 3 of the GNU Lesser General Public License as published
 * by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:  Gary Wang  <gary.wang@canonical.com>
 */

#ifndef MCLOUD_API_CLIENT_H_
#define MCLOUD_API_CLIENT_H_

#include <mcloud/api/visibility.h>
#include <mcloud/api/taskqueue.h>

#include <deque>
#include <future>
#include <string>
#include <thread>
#include <vector>

namespace mcloud {
namespace api {

class CloudResource;
class DownloadTask;
class UploadTask;
class DiskInfo;
class Outlink;
class ClientPriv;

/*!
    \class Client
    \brief Mcloud API provides cmcc users to easily access their cloud contents storaged in mcloud service,
    E.g music, movie, docs, images, and so on.

    The API enable developers to retrieve authorized user cloud content list, cloud storage infomation,
    create and delete cloud resources, create an extranet link to sharing. It is constructed around one sync up manager,
    under the hood, which creates two thread for content uploading/downloading asynchronously.
    Application developer can call \sa add_download_tasks() by passing content id for uploading,
    \sa add_upload_tasks() by passing local file path for downloading.

    \note Mcloud API covers cloud resourece access and content sync-up between local and cloud. It does not take
    responsibility of mcloud authentication. So before using mcloud content APIs, developer need to pass
    user authorization information via \sa refresh_token
 */

class MCLOUD_API_DLL_PUBLIC Client {

  public:
    /*!
     * \brief The TaskStatus enum indicates current status this task is running.
     */
    enum class TaskStatus {
        Unstart,        ///< Task is in queue and sync manager does not run this task yet
        Running,        ///< Task is being processed by sync manager
        Canceled,       ///< Task is in canceled status, it is caused by the call from Client::cancel_sync()
        Paused,         ///< Task is in paused status, it is caused by the call from Client::pause_sync()
        Broken,         ///< Task is broken when sync manager failed to download/upload a task after several times.
        Complete,       ///< Task is completed
    };

    /*!
     * \brief All Content type .
     */
    enum class ContentType {
        all,        ///< all kinds of content type
        image,      ///< png, gif, jpg content
        audio,      ///< mp3, ogg, wav content
        video,      ///< mp4, avi, rmvb, mkv content
        other,      ///< except explicit declared content type in this enum
        doc,        ///< doc, dox content
        speadsheet, ///< xls, xlsx content
        ppt         ///< ppt content
    };

    typedef std::shared_ptr<Client> Ptr;

    typedef std::shared_ptr<DiskInfo> DiskInfoPtr;

    typedef std::shared_ptr<CloudResource>  CloudResourcePtr;
    typedef std::deque<CloudResourcePtr>   ResourceList;

    typedef std::shared_ptr<Outlink> OutlinkPtr;
    typedef std::deque<OutlinkPtr>   OutlinkList;

    typedef std::shared_ptr<DownloadTask> DownloadTaskPtr;
    typedef TaskQueue<DownloadTaskPtr> DownloadList;

    typedef std::shared_ptr<UploadTask> UploadTaskPtr;
    typedef TaskQueue<UploadTaskPtr>  UploadList;

    typedef std::vector<std::string>    Stringlist;

    /*!
     * \brief Constructs a Mcloud Client object with \a request_timeout seconds.
     * Meanwhile it launches sync-up manager to deal with download and upload task by default.
     */
    Client(int request_timeout = 10);

    virtual ~Client();

    /*!
     * \brief Refresh access token by using \a refresh_token.
     * The refresh token can be retrieved from ubuntu sso services(mcloud).
     * \throw std::domain_error if error occurs.
     * \return true if access token is refreshed, otherwise returns false.
     */
    bool refresh_token(const std::string &refresh_token);

    /*!
     * \brief Returnes login user's clcoud storage info.
     * It will be nullptr if no disk info found or error occurs.
     * \throw std::domain_error if error occurs.
     * @sa DiskInfo
     */
    DiskInfoPtr disk_info();

    /*!
     * \brief Returnes sync-up folder id.
     * \throw std::domain_error if error occurs.
     * \note there're a few folders in a user's mcloud root directory, but only one can be used for
     * content sync-up by third-party client.
     * \sa cloud_content_list()
     */
    std::string cloud_sync_folder_id();

    /*!
     * \brief Returns true if the file \a file_path exists on cloud, otherwise return false
     * \throw std::domain_error if error occurs.
     */
    bool exist_on_cloud(const std::string &file_path);

    /*!
     * \brief Returns content list under one specific folder \a folder_id with given type \a content_type.
     * The fetched content count is 200 at most per call. \a content_type is ContentType::all by default.
     * For big content list, \a start_index and \a count can be helpful to retrieve whole content list
     * \throw std::domain_error if error occurs.
     * \sa Client::ContentType
     */
    ResourceList cloud_content_list(int start_index,
                                    int count,
                                    ContentType content_type = ContentType::all,
                                    const std::string &folder_id = std::string());

    /*!
     * \brief Returns true if folders \a folder_ids are deleted, otherwise return false.
     * \throw std::domain_error if error occurs.
     * \sa delete_contents()
     */
    bool delete_folders(const Stringlist &folder_ids);

    /*!
     * \brief Returns true if contents \a content_ids are deleted, otherwise return false.
     * \throw std::domain_error if error occurs.
     * \sa delete_folders()
     */
    bool delete_contents(const Stringlist &content_ids);

    /*!
     * \brief Create folders extranet link with given \a folder_ids and return a list of Outlink
     * \throw std::domain_error if error occurs.
     * \sa create_content_extranet_link()
     */
    OutlinkList create_folder_extranet_link(const Stringlist &folder_ids);

    /*!
     * \brief Create content_extranet link with given \a content_ids and return a list of Outlink
     * \throw std::domain_error if error occurs.
     * \sa create_folder_extranet_link()
     */
    OutlinkList create_content_extranet_link(const Stringlist &content_ids);

    /*!
     * \brief Cancel download or upload task in sync-up thread if sync-up service is running.
     * \throw std::domain_error if error occurs.
     * \note A canceled download or upload task has no chance to run again by sync-up thread
     * \sa start_sync(), pause_sync()
     */
    void cancel_sync();

    /*!
     * \brief Start running download start or upload task in sync-up thread.
     * Sync-up thread will continue to run after it's called if sync-up thread is paused or canceled before
     * \throw std::domain_error if error occurs.
     * \sa pause_sync(), cancel_sync()
     */
    void start_sync();

    /*!
     * \brief Pause download or upload task in sync-up thread if sync-up service is running
     * Sync-up thread can continue to run previous paused task by call \sa start_sync()
     * \throw std::domain_error if error occurs.
     * \sa start_sync(), cancel_sync()
     */
    void pause_sync();

    /*!
     * \brief Push content id list \a contentId_list into download queue.
     * The list is used by sync-up manager to schedule download task automatically.
     * \return download task list that are pushed in sync-up manager.
     * \sa add_download_tasks()
     */
    Client::DownloadList add_download_tasks(const Stringlist &contentId_list);

    /*!
     * \brief Push local files path list \a file_list into upload queue.
     * The list is used by sync-up manager to schedule upload task automatically.
     * Sync manager will detect if the uploading file is already existing on cloud, if so, it simply set the
     * status of upload task completed to avoid duplicate upload
     * \return upload task list that are pushed in sync-up manager.
     * \note A file name can not contain any of following characters:
     *      '\', '/', ':', '*', '?', '"', '<', '>', '|'
     * \sa add_download_tasks()
     */
    Client::UploadList add_upload_tasks(const Stringlist &file_list);

    /*!
     * \brief Returns download task list hold by sync up manager. The list is used by
     * sync-up manager for content download.Each item in the list contains basic information of regarding task.
     * \throw std::domain_error if error occurs.
     * \sa upload_queue(), DownloadTask
     */
    DownloadList download_queue();

    /*!
     * \brief Returns upload task list hold by sync up manager. The list is used by sync-up manager
     * for content upload.Each item in the list contains basic information of regarding task.
     * \throw std::domain_error if error occurs.
     * \sa download_queue(), UploadTask
     */
    UploadList upload_queue();

  private:
    friend ClientPriv;
    std::shared_ptr<ClientPriv>  p;

};

}
}

#endif // MCLOUD_API_CLIENT_H_
