/*
 * Copyright (C) 2016 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of version 3 of the GNU Lesser General Public License as published
 * by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Gary Wang  <gary.wang@canonical.com>
 */

#ifndef MCLOUD_API_DOWNLOADTASK_H_
#define MCLOUD_API_DOWNLOADTASK_H_

#include <mcloud/api/client.h>
#include <mcloud/api/visibility.h>

#include <memory>
#include <string>
#include <atomic>
#include <fstream>

#include <tinyxml2.h>

namespace tinyxml2 {
    class XMLElement;
}

namespace mcloud {
namespace api {

class TaskHandler;

/*!
    \class DownloadTask
    \brief DownloadTask is a task item that can be access from application layer to fetch
    basic download item information also it's used by internal sync-up manager for content download
 */

class MCLOUD_API_DLL_PUBLIC DownloadTask {
public:
    typedef std::function<void(float)> ProgressHandler;

    typedef std::function<void(Client::TaskStatus)> StatusHandler;

    virtual ~DownloadTask() = default;

    /*!
     * \brief Returns an unique id of cloud content in mcloud.
     */
    const std::string content_id() const;

    /*!
     * \brief Returns an display name of cloud content in mcloud.
     */
    const std::string content_name() const;

    /*!
     * \brief Returns cloud content local storage file path.
     */
    const std::string file_path() const;

    /*!
     * \brief Returns a templary item download url assigned by mcloud service.
     * \note this task url will be expried after a period of time
     */
    const std::string task_url() const;

    /*!
     * \brief Returns current download item sync-up status.
     * \sa Client::TaskStatus
     */
    Client::TaskStatus status() const;

    /*!
     * \brief Handler for download progress of a task.
     * \sa ProgressHandler
     */
    DownloadTask::ProgressHandler & progress_changed();

    /*!
     * \brief Handler for download progress of a task.
     * \sa StatusHandler
     */
    DownloadTask::StatusHandler & status_changed();

private:
    DownloadTask(const tinyxml2::XMLElement *root);

    std::shared_ptr<TaskHandler> task_handler();

    template<typename T>
    friend class SyncThread;
    friend class ClientPriv;

    class Priv;
    std::shared_ptr<Priv>   p;
};

}
}

#endif // MCLOUD_API_DOWNLOADTASK_H_
