/*
 * Copyright (C) 2016 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of version 3 of the GNU Lesser General Public License as published
 * by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Gary Wang  <gary.wang@canonical.com>
 */

#ifndef MCLOUD_API_UPLOADTASK_H_
#define MCLOUD_API_UPLOADTASK_H_

#include <mcloud/api/client.h>
#include <mcloud/api/visibility.h>

#include <memory>
#include <string>
#include <atomic>
#include <functional>
#include <fstream>

namespace tinyxml2 {
    class XMLElement;
}

namespace mcloud {
namespace api {

class TaskHandler;

/*!
    \class UploadTask
    \brief UploadTask is a task item that can be access from application layer to fetch
    basic upload item information also it's used by internal sync-up manager for content upload.
 */

class MCLOUD_API_DLL_PUBLIC UploadTask {
public:
    typedef std::function<void(float)> ProgressHandler;

    typedef std::function<void(Client::TaskStatus)> StatusHandler;

    virtual ~UploadTask() = default;

    /*!
     * \brief Returns a unique id of upload task assigned by mcloud.
     */
    const std::string upload_task_id() const;

    /*!
     * \brief Returns a unique id of cloud content in mcloud.
     */
    const std::string content_id() const ;

    /*!
     * \brief Returns a display name of local content.
     */
    const std::string content_name() const;

    /*!
     * \brief Returns uploaded content local storage file path.
     */
    const std::string file_path() const ;

    /*!
     * \brief Returns a item upload url assigned by mcloud service.
     * \note this task url will be expired after a period of time
     */
    const std::string task_url() const ;

    /*!
     * \brief Returns current upload item sync-up status.
     * \sa Client::TaskStatus
     */
    Client::TaskStatus status() const;

    /*!
     * \brief Returns current upload progress of a item in percent.
     */
    float progress() const ;

    /*!
     * \brief Returns file size of upload content.
     */
    size_t file_size() const;

    /*!
     * \brief Returns true if a uploaded item does not exist in cloud, otherwise return false.
     */
    bool is_need_upload() const;

    /*!
     * \brief Handler for upload progress of a task.
     * \sa ProgressHandler
     */
    UploadTask::ProgressHandler & progress_changed();

    /*!
     * \brief Handler for upload status of a task.
     * \sa StatusHandler
     */
    UploadTask::StatusHandler & status_changed();

private:
    UploadTask(const tinyxml2::XMLElement *root,
               const std::string &file_path,
               const std::string &task_id,
               const std::string &task_url);

    std::shared_ptr<TaskHandler> task_handler();

    template<typename T>
    friend class SyncThread;
    friend class ClientPriv;

    class Priv;
    std::shared_ptr<Priv>   p;
};

}
}

#endif // MCLOUD_API_UPLOADTASK_H_
