/*
 * Copyright (C) 2014 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of version 3 of the GNU Lesser General Public License as published
 * by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:  Gary Wang  <gary.wang@canonical.com>
 */

#include <boost/archive/iterators/base64_from_binary.hpp>
#include <boost/archive/iterators/binary_from_base64.hpp>
#include <boost/archive/iterators/transform_width.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
#include <boost/uuid/sha1.hpp>
#include <boost/regex.hpp>
#include <core/net/error.h>
#include <core/net/uri.h>
#include <core/net/http/client.h>
#include <core/net/http/request.h>
#include <core/net/http/response.h>
#include <tinyxml2.h>
#include <iostream>

#include <mcloud/api/cloudfolder.h>
#include <mcloud/api/cloudcontent.h>
#include <mcloud/api/diskinfo.h>
#include <mcloud/api/downloadtask.h>
#include <mcloud/api/uploadtask.h>
#include <mcloud/api/cloudresource.h>
#include <mcloud/api/outlink.h>

#include <openssl/md5.h>
#include <iomanip>

#include "client_priv.h"
#include "syncmanager.h"
#include "config.h"

namespace http = core::net::http;
namespace net = core::net;

namespace fs = boost::filesystem;
namespace alg = boost::algorithm;

using namespace mcloud::api;
using namespace std;

static const std::string REGISTER_APP_NAME = "和彩云ubuntu";
static const std::string THIRDPARTY_ANONYMOUS_ACCOUNT = "thirdparty_anonymous_account";

namespace {

void log_ouput(const std::string &info) {
#ifndef NDEBUG
    cout << info << endl;
#else
    (void)(info);
#endif
}

std::string encode64(const string& val) {
    using namespace boost::archive::iterators;
    typedef base64_from_binary<transform_width<std::string::const_iterator, 6, 8>> iter;
    auto tmp = std::string(iter(std::begin(val)), iter(std::end(val)));
    return tmp.append((3 - val.size() % 3) % 3, '=');
}

static std::string md5sum(std::string filepath) {
    std::ifstream  ifs(filepath, std::ios::binary);

    if (!ifs.good()) {
        std::cerr << "file is invalid: " << filepath << std::endl;
        return std::string();
    }

    unsigned char hash[MD5_DIGEST_LENGTH];
    MD5_CTX mdContext;
    char data[1024];

    MD5_Init(&mdContext);
    while(ifs.good()) {
        ifs.read(data, sizeof(data));
        MD5_Update (&mdContext, data, ifs.gcount());
    }
    MD5_Final (hash,&mdContext);

    std::stringstream stream;
    for(int i = 0; i < MD5_DIGEST_LENGTH; i++) {
        stream << std::hex << std::setw(2) << std::setfill('0') << (int)hash[i];
    }

    return stream.str();
}

static void append_child(tinyxml2::XMLPrinter &printer,
                         const std::string & field,
                         const std::string & value) {
    printer.OpenElement(field.c_str());
    printer.PushText(value.c_str());
    printer.CloseElement();
}

static string generate_diskinfo_body() {
    tinyxml2::XMLPrinter printer;
    printer.OpenElement("getDiskInfo");
    append_child(printer, "MSISDN", THIRDPARTY_ANONYMOUS_ACCOUNT.c_str());
    printer.CloseElement();

    return string(printer.CStr());
}

static string generate_content_list_body(int start_index = -1,
                                         int count = 200,
                                         Client::ContentType content_type = Client::ContentType::all,
                                         const std::string & shared_folder_id = std::string()) {
    tinyxml2::XMLPrinter printer;
    printer.OpenElement("getDisk");

    append_child(printer, "MSISDN", THIRDPARTY_ANONYMOUS_ACCOUNT.c_str());
    append_child(printer, "catalogID", shared_folder_id.empty() ? "root" : shared_folder_id.c_str());
    append_child(printer, "filterType", "0");
    append_child(printer, "catalogSortType", "0");
    append_child(printer, "contentType", std::to_string((int)content_type).c_str());
    append_child(printer, "contentSortType", "0");
    append_child(printer, "sortDirection", "1");
    append_child(printer, "startNumber",std::to_string(start_index).c_str());
    append_child(printer, "endNumber",  std::to_string(start_index + count).c_str());
    append_child(printer, "catalogType", "-1");

    printer.CloseElement();

    return string(printer.CStr());
}

static string generate_delete_body(const std::vector<std::string> &dir_ids,
                                   const std::vector<std::string> &content_ids) {
    tinyxml2::XMLPrinter printer;
    printer.OpenElement("delCatalogContent");

    append_child(printer, "MSISDN", THIRDPARTY_ANONYMOUS_ACCOUNT.c_str());
    printer.OpenElement("catalogIDs");
    printer.PushAttribute("length", std::to_string(dir_ids.size()).c_str());
    for (auto & dir_id: dir_ids) {
        append_child(printer, "ID", dir_id.c_str());
    }
    printer.CloseElement();

    printer.OpenElement("contentIDs");
    printer.PushAttribute("length", std::to_string(content_ids.size()).c_str());
    for (auto & content_id: content_ids) {
        append_child(printer, "ID", content_id.c_str());
    }
    printer.CloseElement();

    printer.CloseElement();

    return string(printer.CStr());
}

static string generate_extranet_body(const std::vector<std::string> &dir_ids,
                                     const std::vector<std::string> &content_ids) {
    tinyxml2::XMLPrinter printer;
    printer.OpenElement("getOutLink");
    printer.OpenElement("getOutLinkReq");

    append_child(printer, "account", THIRDPARTY_ANONYMOUS_ACCOUNT.c_str());
    printer.OpenElement("caIDLst");
    printer.PushAttribute("length", std::to_string(dir_ids.size()).c_str());
    for (auto & dir_id: dir_ids) {
        append_child(printer, "item", dir_id.c_str());
    }
    printer.CloseElement();

    printer.OpenElement("coIDLst");
    printer.PushAttribute("length", std::to_string(content_ids.size()).c_str());
    for (auto & content_id: content_ids) {
        append_child(printer, "item", content_id.c_str());
    }
    printer.CloseElement();

    append_child(printer, "pubType", "1");

    printer.CloseElement();
    printer.CloseElement();

    return string(printer.CStr());
}

static string generate_download_body(const std::vector<std::string> &content_ids) {
    tinyxml2::XMLPrinter printer;
    printer.OpenElement("downloadRequest");
    append_child(printer, "appName", REGISTER_APP_NAME.c_str());
    append_child(printer, "MSISDN", THIRDPARTY_ANONYMOUS_ACCOUNT.c_str());

    string content_list = alg::join(content_ids, "|");

    append_child(printer, "contentID", content_list.c_str());
    append_child(printer, "OwnerMSISDN", "");
    append_child(printer, "entryShareCatalogID", "");
    append_child(printer, "operation", "0");
    append_child(printer, "fileVersion", "-1");

    printer.CloseElement();

    return string(printer.CStr());
}

static string generate_upload_body(const std::vector<std::string> &file_list,
                                   const std::string folder_id = std::string()) {
    tinyxml2::XMLPrinter printer;
    printer.OpenElement("pcUploadFileRequest");
    append_child(printer, "ownerMSISDN", THIRDPARTY_ANONYMOUS_ACCOUNT.c_str());
    append_child(printer, "fileCount", std::to_string(file_list.size()).c_str());

    //total_size filed must be decleared firstly
    int64_t file_size = 0;
    for (auto & file_path : file_list) {
        if (fs::exists(file_path)) {
            file_size += fs::file_size(file_path);
        }
    }
    append_child(printer, "totalSize", std::to_string(file_size).c_str());

    printer.OpenElement("uploadContentList");
    printer.PushAttribute("length", std::to_string(file_list.size()).c_str());

    for (auto & file_path : file_list) {
        std::string hash = md5sum(file_path);

        file_size = 0;
        if (fs::exists(file_path)) {
            file_size = fs::file_size(file_path);
        }

        printer.OpenElement("uploadContentInfo");
        append_child(printer, "contentName", fs::path(file_path).filename().c_str());
        append_child(printer, "contentSize", std::to_string(file_size).c_str());
        append_child(printer, "contentDesc", "");
        append_child(printer, "contentTAGList", "");
        append_child(printer, "digest", hash);

        printer.CloseElement();
    }

    printer.CloseElement();

    append_child(printer, "parentCatalogID", folder_id);
    printer.CloseElement();

    return string(printer.CStr());
}
}

class ClientPriv::HttpClient {
public:
    HttpClient(int request_timeout) :
        client_(http::make_client()),
        worker_ { [this]() {client_->run();} },
        cancelled_(false) {
        config_.request_timeout = request_timeout;
    }

    ~HttpClient() {
        client_->stop();
        if (worker_.joinable()) {
            worker_.join();
        }
    }

    std::shared_ptr<core::net::http::Client> client_;

    std::thread worker_;

    Config config_;

    std::mutex config_mutex_;

    std::atomic<bool> cancelled_;

    void get(const net::Uri::Path &path,
            const net::Uri::QueryParameters &parameters,
            http::Request::Handler &handler) {
        auto configuration = net_config(path, parameters);
        configuration.header.add("User-Agent", config_.user_agent);

        auto request = client_->head(configuration);
        request->async_execute(handler);
    }

    void post(const net::Uri::Path &path,
            const net::Uri::QueryParameters &parameters,
            const std::string &postbody,
            const std::string &content_type,
            http::Request::Handler &handler) {
        std::lock_guard<std::mutex> lock(config_mutex_);
        http::Request::Configuration configuration = net_config(path, parameters);
        configuration.header.add("User-Agent", config_.user_agent);
        configuration.header.add("Content-Length", std::to_string(postbody.size()));
        configuration.header.add("Content-Type", content_type);

        auto request = client_->post(configuration, postbody, content_type);
        request->async_execute(handler);
    }

    http::Request::Configuration net_config(const net::Uri::Path &path,
                                            const net::Uri::QueryParameters &parameters) {
        http::Request::Configuration configuration;
        net::Uri::QueryParameters complete_parameters(parameters);
        if (!config_.access_token.empty()) {
            configuration.header.add("Authorization",
                    "Bearer " + config_.access_token);
        } else {
            configuration.header.add("Authorization",
                    "Basic " + encode64(config_.client_id + ":" + config_.client_password));
        }

        if (getenv("MCLOUD_LOCAL_SERVER_URL")) {
            config_.apiroot = string(getenv("MCLOUD_LOCAL_SERVER_URL"));
        }

        net::Uri uri = net::make_uri(config_.apiroot, path,
                complete_parameters);
        configuration.uri = client_->uri_to_string(uri);

        return configuration;
    }

    http::Request::Progress::Next progress_changed(
            const http::Request::Progress&) {
        return cancelled_ ?
                http::Request::Progress::Next::abort_operation :
                http::Request::Progress::Next::continue_operation;
    }

    template<typename T>
    future<T> async_get(const net::Uri::Path &path,
            const net::Uri::QueryParameters &parameters,
            const function<T(const tinyxml2::XMLElement *root)> &func) {
        auto prom = make_shared<promise<T>>();

        http::Request::Handler handler;
        handler.on_progress(
                bind(&ClientPriv::HttpClient::progress_changed, this, placeholders::_1));
        handler.on_error([prom](const net::Error& e)
        {
            prom->set_exception(make_exception_ptr(domain_error(e.what())));
        });
        handler.on_response(
            [prom,func](const http::Response& response)
            {
                tinyxml2::XMLDocument doc;
                tinyxml2::XMLError error = doc.Parse(response.body.c_str());
                const tinyxml2::XMLElement *root = doc.FirstChildElement();

                if (error != tinyxml2::XML_SUCCESS) {
                    prom->set_exception(make_exception_ptr(domain_error("xml parse failed")));
                } else if (response.status != http::Status::ok) {
                    prom->set_exception(make_exception_ptr(domain_error(root->GetText())));
                } else {
                    prom->set_value(func(root));
                }
            }
        );

        get(path, parameters, handler);

        return prom->get_future();
    }

    template<typename T>
    future<T> async_post(const net::Uri::Path &path,
            const net::Uri::QueryParameters &parameters,
            const std::string &postmsg,
            const std::string &content_type,
            const function<T(const tinyxml2::XMLElement *root)> &func) {
        auto prom = make_shared<promise<T>>();

        http::Request::Handler handler;
        handler.on_progress(
                bind(&ClientPriv::HttpClient::progress_changed, this, placeholders::_1));
        handler.on_error([prom](const net::Error& e)
        {
            prom->set_exception(make_exception_ptr(domain_error(e.what())));
        });
        handler.on_response(
            [path, prom,func](const http::Response& response)
            {
                std::vector<string>::const_iterator iter =
                        std::find_if(path.begin(), path.end(), [=](const string & pathitem){
                    return (pathitem == "OAuth2");
                });

                //For authentication, response is json-based data
                //create a xml-based data,
                tinyxml2::XMLDocument doc;
                tinyxml2::XMLError error = tinyxml2::XML_NO_ERROR;
                if (iter != path.end()) {
                    std::string xml_response = "<result resultCode=\"0\"><![CDATA[" + response.body + "]]></result>";
                    error = doc.Parse(xml_response.c_str());
                } else {
                    error = doc.Parse(response.body.c_str());
                }
                const tinyxml2::XMLElement *result = doc.FirstChildElement();

                if (error != tinyxml2::XML_SUCCESS) {
                    prom->set_exception(make_exception_ptr(domain_error("xml parse failed")));
                } else if ((response.status != http::Status::created &&
                       response.status != http::Status::ok &&
                       response.status != http::Status::no_content ) ||
                       strcmp(result->Attribute("resultCode"), "0") != 0) {
                    prom->set_exception(make_exception_ptr(domain_error(result->Attribute("desc"))));
                } else {
                    prom->set_value(func(result));
                }
            }
        );

        post(path, parameters, postmsg, content_type, handler);

        return prom->get_future();
    }
};

ClientPriv::ClientPriv(int request_timeout)
    : client_(make_shared<HttpClient>(request_timeout)),
      sync_manager_(make_shared<SyncManager>()){
    start_sync();
}

ClientPriv::~ClientPriv() {
    client_->cancelled_ = true;
}

string ClientPriv::cloud_sync_folder_id() {
    Client::ResourceList folderlist = cloud_root_folder_list();
    for (auto &folder: folderlist) {
        mcloud::api::CloudFolder::Ptr cloudfolder = std::static_pointer_cast<mcloud::api::CloudFolder>(folder);
        if (cloudfolder->folder_type() == mcloud::api::CloudFolder::FolderType::sync) {
            return cloudfolder->id();
        }
    }

    return std::string();
}

bool ClientPriv::exist_on_cloud(const std::string &file_path) {
    std::vector<std::string> file_path_list(1, file_path);
    string postbody = generate_upload_body(file_path_list);
    string content_type = "text/xml;UTF-8";

    log_ouput(postbody);

    auto upload_future = client_->async_post<bool>(
            { "richlifeApp", "devapp", "IUploadAndDownload" }, { }, postbody, content_type,
            [](const tinyxml2::XMLElement *root) {
        bool is_exists = false;

        auto upload_ele = root->FirstChildElement("uploadResult");
        auto new_content_list = upload_ele->FirstChildElement("newContentIDList");
        if (new_content_list) {
            auto new_content_ele = new_content_list->FirstChildElement();
            if(new_content_ele) {
                is_exists = strcmp(new_content_ele->FirstChildElement("isNeedUpload")->GetText(), "1");
            }
        }

        return is_exists;
    });

    return get_or_throw(upload_future);
}

DiskInfo::Ptr ClientPriv::disk_info() {
    std::string postbody = generate_diskinfo_body();
    std::string content_type = "text/xml;UTF-8";

    log_ouput(postbody);

    auto disk_future = client_->async_post<DiskInfo::Ptr>(
    { "richlifeApp", "devapp", "IUser" }, { }, postbody, content_type,
                [](const tinyxml2::XMLElement *root) {
        return std::shared_ptr<DiskInfo>(new DiskInfo(root));
    });

    return get_or_throw(disk_future);
}

Client::ResourceList ClientPriv::cloud_content_list(int start_index,
                                                int count,
                                                Client::ContentType type,
                                                const std::string & folder_id) {
    std::string postbody = generate_content_list_body(start_index,
                                                      count,
                                                      type,
                                                      folder_id);
    std::string content_type = "text/xml;UTF-8";

    log_ouput(postbody);

    auto resource_future = client_->async_post<Client::ResourceList>(
            { "richlifeApp", "devapp", "ICatalog" }, { }, postbody, content_type,
            [](const tinyxml2::XMLElement *root) {
        auto res = root->FirstChildElement();

        Client::ResourceList results;

        auto cata_folders_ele = res->FirstChildElement("catalogList");
        if (cata_folders_ele) {
            auto folder_ele = cata_folders_ele->FirstChildElement();
            while(folder_ele) {
                results.emplace_back(std::shared_ptr<CloudFolder>(
                                         new CloudFolder(folder_ele)));
                folder_ele = folder_ele->NextSiblingElement();
            }
        }

        auto cata_contents_ele = res->FirstChildElement("contentList");
        if (cata_contents_ele) {
            auto content_ele = cata_contents_ele->FirstChildElement();
            while(content_ele) {
                results.emplace_back(std::shared_ptr<CloudContent>(
                                         new CloudContent(content_ele)));
                content_ele = content_ele->NextSiblingElement();
            }
        }

        return results;
    });

    return get_or_throw(resource_future);
}

bool ClientPriv::delete_folders(const Client::Stringlist &folder_ids) {
    string postbody = generate_delete_body(folder_ids, vector<string>{});
    return delete_internal(postbody);
}

bool ClientPriv::delete_contents(const Client::Stringlist &content_ids) {
    string postbody = generate_delete_body(vector<string>{}, content_ids);
    return delete_internal(postbody);
}

Client::OutlinkList ClientPriv::create_folder_extranet_link(
        const Client::Stringlist &folder_ids) {
    string postbody = generate_extranet_body(folder_ids, Client::Stringlist{});
    return create_extranet_internal(postbody);
}

Client::OutlinkList ClientPriv::create_content_extranet_link(
        const Client::Stringlist &content_ids) {
    string postbody = generate_extranet_body(Client::Stringlist{}, content_ids);
    return create_extranet_internal(postbody);
}

Client::DownloadList ClientPriv::create_download_link(
        const Client::Stringlist &content_ids) {
    string postbody = generate_download_body(content_ids);
    string content_type = "text/xml;UTF-8";

    log_ouput(postbody);

    auto download_future = client_->async_post<Client::DownloadList>(
            { "richlifeApp", "devapp", "IUploadAndDownload" }, { }, postbody, content_type,
            [](const tinyxml2::XMLElement *root) {
        Client::DownloadList results;

        auto link_set_ele = root->FirstChildElement("String");
        if (link_set_ele) {
            auto current_link_ele = link_set_ele->FirstChildElement();
            while(current_link_ele) {
                results.push(std::shared_ptr<DownloadTask>(new DownloadTask(current_link_ele)));
                current_link_ele = current_link_ele->NextSiblingElement();
            }

            if (link_set_ele->GetText()) {
                results.push(std::shared_ptr<DownloadTask>(new DownloadTask(link_set_ele)));
            }
        }

        return results;
    });

    return get_or_throw(download_future);
}

Client::UploadList ClientPriv::create_upload_link(const Client::Stringlist &file_paths,
                                              const string &cloud_folder_id) {
    string postbody = generate_upload_body(file_paths, cloud_folder_id);
    string content_type = "text/xml;UTF-8";

    log_ouput(postbody);

    auto upload_future = client_->async_post<Client::UploadList>(
            { "richlifeApp", "devapp", "IUploadAndDownload" }, { }, postbody, content_type,
            [&file_paths](const tinyxml2::XMLElement *root) {
        Client::UploadList results;

        auto upload_ele = root->FirstChildElement("uploadResult");
        std::string task_id = upload_ele->FirstChildElement("uploadTaskID")->GetText();

        std::string redirection_url;
        if (upload_ele->FirstChildElement("redirectionUrl")) {
            redirection_url = upload_ele->FirstChildElement("redirectionUrl")->GetText();
        }

        auto new_content_list = upload_ele->FirstChildElement("newContentIDList");
        if (new_content_list) {
            auto new_content_ele = new_content_list->FirstChildElement();
            while(new_content_ele) {
                auto content_name = new_content_ele->FirstChildElement("contentName")->GetText();

                std::vector<std::string>::const_iterator it = std::find_if(file_paths.begin(), file_paths.end(),
                                                           [&content_name](const std::string& path) -> bool {
                    return (fs::path(path).filename() == content_name);
                });

                if (it != file_paths.end()) {
                    results.push(std::shared_ptr<UploadTask>(new UploadTask(new_content_ele,
                                                        *it,
                                                        task_id,
                                                        redirection_url)));
                }
                new_content_ele = new_content_ele->NextSiblingElement();
            }
        }

        return results;
    });

    return get_or_throw(upload_future);
}

Client::ResourceList ClientPriv::cloud_root_folder_list() {
    std::string postbody = generate_content_list_body();
    std::string content_type = "text/xml;UTF-8";

    log_ouput(postbody);

    auto resource_future = client_->async_post<Client::ResourceList>(
            { "richlifeApp", "devapp", "ICatalog" }, { }, postbody, content_type,
            [](const tinyxml2::XMLElement *res) {
        auto root = res->FirstChildElement();
        Client::ResourceList results;

        auto cata_folders_ele = root->FirstChildElement("catalogList");
        if (cata_folders_ele) {
            auto folder_ele = cata_folders_ele->FirstChildElement();
            while(folder_ele) {
                results.emplace_back(std::shared_ptr<CloudFolder>(new CloudFolder(folder_ele)));
                folder_ele = folder_ele->NextSiblingElement();
            }
        }

        return  results;
    });

    return get_or_throw(resource_future);
}

bool ClientPriv::delete_internal(const string &postbody) {
    string content_type = "text/xml;UTF-8";

    log_ouput(postbody);

    auto delete_future =  client_->async_post<bool>(
            { "richlifeApp", "devapp", "IContent" }, { }, postbody, content_type,
            [](const tinyxml2::XMLElement *root) {
                return !strcmp(root->Attribute("resultCode"), "0");
    });

    return get_or_throw(delete_future);
}

Client::OutlinkList ClientPriv::create_extranet_internal(const string &postbody) {
    string content_type = "text/xml;UTF-8";

    log_ouput(postbody);

    auto create_future = client_->async_post<Client::OutlinkList>(
            { "richlifeApp", "devapp", "IOutLink" }, { }, postbody, content_type,
            [](const tinyxml2::XMLElement *root) {
        Client::OutlinkList results;

        auto link_set_ele = root->FirstChildElement("getOutLinkRes")->FirstChildElement("getOutLinkResSet");
        if (link_set_ele) {
            auto current_link_ele = link_set_ele->FirstChildElement();
            while(current_link_ele) {
                results.emplace_back(std::shared_ptr<Outlink>(new Outlink(current_link_ele)));
                current_link_ele = current_link_ele->NextSiblingElement();
            }
        }

        return results;
    });

    return get_or_throw(create_future);
}

bool ClientPriv::refersh_token(const string &refresh_token) {
    const std::string content_type = "application/x-www-form-urlencoded";
    const std::string postbody = "grant_type=refresh_token&refresh_token="
                                    + refresh_token;
    log_ouput(postbody);

    auto refresh_token_future = client_->async_post<bool>(
            { "oauthApp", "OAuth2", "refreshToken" }, { }, postbody, content_type,
            [this](const tinyxml2::XMLElement *root) {
        boost::cmatch cap;
        const boost::regex exp(".*\"(.*?)\"\\}");
        if (boost::regex_match(root->GetText(), cap, exp)) {
            client_->config_.access_token = cap[1];
            return true;
        }
        return false;
    });

    return get_or_throw(refresh_token_future);
}

void ClientPriv::cancel_sync() {
    sync_manager_->cancel();
}

void ClientPriv::start_sync() {
    sync_manager_->start();
}

void ClientPriv::pause_sync() {
    sync_manager_->pause();
}

Client::DownloadList ClientPriv::add_download_tasks(const Client::Stringlist &contentId_list) {
    //Only support to create single content item since
    //The NDA we are granted cannot create link with multi resource
    //when the configuration value of 'actBatchOutlink' is off.
    //so compose new vector and pass content id one by one

    Client::DownloadList tasklist;

    for (auto & content_id : contentId_list) {
        vector<string> contents{content_id};

        //If content id is invalid e.g. content id doesn't exist or
        //auth user is not the owner of content, just skip it
        //let sync mananger continue execute another task
        try {
            Client::DownloadList items = create_download_link(contents);
            for (auto &item: items) {
                log_ouput("download link url is: " + item->task_url());
                add_download_task(item);
            }
            tasklist.push(items);
        } catch(domain_error &e) {
            cerr << "sync content downlaod error: " << e.what() << endl;
            continue;
        }
    }

    return tasklist;
}

Client::UploadList ClientPriv::add_upload_tasks(const Client::Stringlist &file_list) {
    //Only support to create single content item since
    //The NDA we are granted cannot create link with multi resource
    //when the configuration value of 'actBatchOutlink' is off.
    //so compose new vector and pass file path one by one

    Client::UploadList tasklist;

    for (auto & file : file_list) {
        vector<string> files{file};

        try {
            Client::UploadList items = create_upload_link(files);
            for (auto &item: items) {
                log_ouput("upload link url is: " + item->task_url());
                add_upload_task(item);
            }
            tasklist.push(items);
        } catch(domain_error &e) {
            cerr << "sync content upload error: " << e.what() << endl;
            continue;
        }
    }

    return tasklist;
}

void ClientPriv::cancel() {
    client_->cancelled_ = true;
}

void ClientPriv::add_download_task(Client::DownloadTaskPtr task) {
    sync_manager_->push_download_task(task);
}

void ClientPriv::add_upload_task(Client::UploadTaskPtr task) {
    sync_manager_->push_upload_task(task);
}

TaskQueue<Client::DownloadTaskPtr> ClientPriv::download_queue() {
    return sync_manager_->download_queue();
}

TaskQueue<Client::UploadTaskPtr> ClientPriv::upload_queue() {
    return sync_manager_->upload_queue();
}

template<typename T> T ClientPriv::get_or_throw(std::future<T> &f) {
    if (f.wait_for(std::chrono::seconds(client_->config_.request_timeout)) != std::future_status::ready) {
        throw std::domain_error("HTTP request timeout");
    }

    return f.get();
}
