/*
 * Copyright © 2016 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Gary Wang <gary.wang@canonical.com>
 */


#include <boost/archive/iterators/base64_from_binary.hpp>
#include <boost/archive/iterators/binary_from_base64.hpp>
#include <boost/archive/iterators/transform_width.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/regex.hpp>
#include <boost/filesystem.hpp>
#include <mcloud/api/downloadtask.h>

#include <iostream>

#include "taskhandler.h"

namespace fs = boost::filesystem;

using namespace mcloud::api;
using namespace std;

static const std::string public_download_dir = std::string(getenv("HOME")) + "/Downloads";

static const int broken_limit = 3;

namespace {
std::string decode64(const std::string &val) {
    using namespace boost::archive::iterators;
    typedef transform_width<binary_from_base64<std::string::const_iterator>, 8, 6> iter;
    return boost::algorithm::trim_right_copy_if(std::string(iter(std::begin(val)), iter(std::end(val))), [](char c) {
        return c == '\0';
    });
}
}

class DownloadTask::Priv {
public:
    Priv(const tinyxml2::XMLElement *root)
        :progress_handler_(nullptr),
         status_handler_(nullptr),
         task_handler_(std::make_shared<TaskHandler>()){
        if (root->FirstChildElement("String"))
            task_url_ = root->FirstChildElement("String")->GetText();
        else
            task_url_ = root->GetText();

        boost::cmatch cap;
        const boost::regex exp(".*contentID=(.*?)&.*filename=(.*?)");

        if (boost::regex_match(task_url_.c_str(), cap, exp)) {
            content_id_ = cap[1];
            content_name_ = decode64(cap[2]);
        }

        // create temp file path in temp path for writing
        file_path_ = fs::path(fs::temp_directory_path() / fs::unique_path())
                                            .string();

        task_handler_->on_prepare() = [this](void*){
            if (ofs_.is_open())
                ofs_.close();

            ofs_.open(file_path_, std::ofstream::out | std::ofstream::trunc);
            if (ofs_.fail()) {
                throw domain_error("failed to open file" + file_path_ + "  error: " + strerror(errno));
            }

            //change status
            task_handler_->set_status(Client::TaskStatus::Running);
        };

        task_handler_->on_progress() = [this](float percent){
            if (progress_handler_) {
                progress_handler_(percent);
            }
        };

        task_handler_->on_status() = [this](Client::TaskStatus status){
            if (status_handler_) {
                status_handler_(status);
            }
        };

        task_handler_->on_finished() = [this]() -> bool {
            ofs_.close();

            std::string download_dir;
            if (getenv("MCLOUD_DOWNLOAD_FOLDER")) {
                download_dir = string(getenv("MCLOUD_DOWNLOAD_FOLDER"));
            } else {
                download_dir = public_download_dir;
            }

            if (!fs::exists(fs::path(download_dir))) {
                try {
                    fs::create_directory(fs::path(download_dir));
                } catch(boost::filesystem::filesystem_error& e) {
                    cerr << "create download directory failed"<< download_dir
                         <<  "error: " << e.what() << endl;
                }
                return false;
            }

            //check if there is a duplicate one
            std::string stem = fs::path(content_name_).stem().string();
            fs::path dest = download_dir + "/" + content_name_;
            int dup_count = 0;
            while (fs::exists(dest)) {
                dest = fs::path(download_dir) / fs::path(stem
                                                            + "_"
                                                            + std::to_string(++dup_count)
                                                            + dest.extension().string());
            }

            //change status
            task_handler_->reset_broken_counter();
            task_handler_->set_status(Client::TaskStatus::Complete);

            try {
                // rename in final when download is complete.
                fs::rename(file_path_, dest);
                file_path_ = dest.string();


            } catch (boost::filesystem::filesystem_error& e) {
                cerr << "create download directory failed"<< download_dir
                     <<  "error: " << e.what() << endl;
                return false;
            }

            return true;
        };

        task_handler_->on_ready_read() = [this](const string &data){
            ofs_.write(data.c_str(), data.size());
        };
    }

    virtual ~Priv() = default;

    std::string task_url_;

    std::string content_id_;

    std::string content_name_;

    std::string file_path_;

    std::ofstream   ofs_;

    DownloadTask::ProgressHandler progress_handler_;

    DownloadTask::StatusHandler status_handler_;

    std::shared_ptr<TaskHandler> task_handler_;
};

DownloadTask::DownloadTask(const tinyxml2::XMLElement *root)
    :p(std::make_shared<DownloadTask::Priv>(root)){
}

std::shared_ptr<TaskHandler> DownloadTask::task_handler() {
    return p->task_handler_;
}

const string DownloadTask::task_url() const {
    return p->task_url_;
}

Client::TaskStatus DownloadTask::status() const {
    return p->task_handler_->status();
}

const string DownloadTask::content_id() const {
    return p->content_id_;
}

const string DownloadTask::content_name() const {
    return p->content_name_;
}

const string DownloadTask::file_path() const {
    return p->file_path_;
}

DownloadTask::ProgressHandler &DownloadTask::progress_changed() {
    return p->progress_handler_;
}

DownloadTask::StatusHandler &DownloadTask::status_changed() {
    return p->status_handler_;
}
