/*
 * Copyright © 2016 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Gary Wang <gary.wang@canonical.com>
 */

#include <boost/algorithm/string.hpp>
#include <core/net/http/streaming_client.h>
#include <core/net/http/streaming_request.h>
#include <core/net/http/response.h>
#include <core/net/http/status.h>

#include <mcloud/api/client.h>
#include <mcloud/api/uploadtask.h>
#include <mcloud/api/downloadtask.h>

#include <fstream>
#include <future>
#include <iostream>

#include "syncthread.h"
#include "taskhandler.h"

namespace http = core::net::http;
namespace net = core::net;

using namespace mcloud::api;
using namespace std;

template <typename T>
class SyncThread<T>::HttpStreamClient {
public:
    HttpStreamClient(int request_timeout) :
        request_timeout_(request_timeout),
        client_(http::make_streaming_client()),
        worker_ { [this]() {client_->run();} }{
    }

    ~HttpStreamClient() {
        client_->stop();
        if (worker_.joinable()) {
            worker_.join();
        }
    }

    void async_get(shared_ptr<T> ptr,
                            const http::Header &headers) {
        auto prom = make_shared<promise<bool>>();

        http::Request::Handler handler;
        handler.on_progress([ptr](const http::Request::Progress& progress)
            {
                if (ptr->task_handler()->status() == Client::TaskStatus::Canceled) {
                    return http::Request::Progress::Next::abort_operation;
                }

                if (progress.download.current > 0.0 &&  progress.download.total > 0.0) {
                    double percent = progress.download.current / progress.download.total;
                    ptr->task_handler()->on_progress()(percent);
                }

                return http::Request::Progress::Next::continue_operation;
            }
        );
        handler.on_error([prom](const net::Error& e)
            {
                prom->set_exception(make_exception_ptr(domain_error(e.what())));
        });
        handler.on_response([prom, ptr](const http::Response& response)
            {
                if (response.status != http::Status::ok) {
                    prom->set_exception(make_exception_ptr(domain_error(response.body)));
                } else {
                    prom->set_value(ptr->task_handler()->on_finished()());
                }
            }
        );

        auto http_config = http::Request::Configuration::from_uri_as_string(
                    ptr->task_url());
        http_config.header = headers;

        std::shared_future<bool> future = prom->get_future();
        auto request = client_->streaming_get(http_config);
        request->set_timeout(std::chrono::milliseconds{request_timeout_});
        request->async_execute(handler, ptr->task_handler()->on_ready_read());

        ptr->task_handler()->attach_request(request, future);
    }

    void async_post(shared_ptr<T> ptr,
                            const http::Header &headers,
                            std::istream &payload,
                            std::size_t size) {
        auto prom = make_shared<promise<bool>>();

        http::Request::Handler handler;
        handler.on_progress([ptr](const http::Request::Progress& progress)
            {
                if (ptr->status() == Client::TaskStatus::Canceled) {
                    return http::Request::Progress::Next::abort_operation;
                }

                if (progress.upload.current > 0.0 &&  progress.upload.total > 0.0) {
                    double percent = (progress.upload.current / progress.upload.total);
                    ptr->task_handler()->on_progress()(percent);
                }

                return http::Request::Progress::Next::continue_operation;
            }
        );
        handler.on_error([prom](const net::Error& e)
            {
                prom->set_exception(make_exception_ptr(domain_error(e.what())));
        });
        handler.on_response([prom, ptr](const http::Response& response)
            {
                if (response.status != http::Status::ok) {
                    prom->set_exception(make_exception_ptr(domain_error(response.body)));
                } else {
                    prom->set_value(ptr->task_handler()->on_finished()());
                }
            }
        );

        auto http_config = http::Request::Configuration::from_uri_as_string(
                    ptr->task_url());
        http_config.header = headers;

        std::shared_future<bool> future = prom->get_future();
        auto request = client_->streaming_post(http_config, payload, size);
        request->set_timeout(std::chrono::milliseconds{request_timeout_});
        request->async_execute(handler, ptr->task_handler()->on_ready_read());

        ptr->task_handler()->attach_request(request, future);
    }

    int request_timeout_;

    std::shared_ptr<core::net::http::StreamingClient> client_;

    std::thread worker_;
};

template<typename T>
SyncThread<T>::SyncThread()
    : client_(std::make_shared<SyncThread::HttpStreamClient>(sync_timeout_ * 500)),
      stop_(true){
}

template <typename T>
SyncThread<T>::~SyncThread() {
    stop();
}

template <typename T>
void SyncThread<T>::start() {
    pause_ = false;

    if (stop_ == true) {
        stop_ = false;
        thread_ = std::thread(&SyncThread::run, this);
    } else {
        //if it's running, then resume
        for (auto & item_ptr : task_queue_) {
            if (item_ptr->status() == Client::TaskStatus::Paused) {
                item_ptr->task_handler()->resume();
            }
        }
    }
}

template <typename T>
void SyncThread<T>::cancel() {
    stop();
}

template <typename T>
void SyncThread<T>::pause() {
    //cancel task if it's running
    for (auto & item_ptr : task_queue_) {
        if (item_ptr->status() == Client::TaskStatus::Running) {
            item_ptr->task_handler()->pause();
        }
    }

    pause_ = true;
}

template <typename T>
void SyncThread<T>::stop() {
    if (stop_)
        return;

    //cancel task if it's running
    for (auto & item_ptr : task_queue_) {
        if (item_ptr->status() == Client::TaskStatus::Running
                || item_ptr->status() == Client::TaskStatus::Paused) {
            item_ptr->task_handler()->cancel();
        }
    }

    //terminate thread
    stop_ = true;
    con_.notify_all();

    if (thread_.joinable())
        thread_.join();
}

template <typename T>
void SyncThread<T>::push_task(std::shared_ptr<T> task_ptr) {
    task_queue_.push(task_ptr);
}

template <typename T>
TaskQueue<std::shared_ptr<T>>  SyncThread<T>::task_queue() {
    return task_queue_;
}

template <typename T>
void SyncThread<T>::run() {
    std::unique_lock<std::mutex> lock(mutex_);

    while(!stop_) {
        con_.wait_for(lock, std::chrono::seconds(sync_timeout_));
        if (!stop_ && !pause_) {
            //check if there is unstared task item in task queue
            typename TaskQueue<std::shared_ptr<T>>::const_iterator it = std::find_if(
                        task_queue_.begin(),
                        task_queue_.end(),
                        [](std::shared_ptr<T> task){
                return (task->status() == Client::TaskStatus::Unstart);
            });

            if (it == task_queue_.end()) {
                continue;
            }

            std::shared_ptr<TaskHandler> task_handler = (*it)->task_handler();

            try {
                task_handler->on_prepare()(nullptr);
            } catch (std::domain_error e) {
                cerr << "error: " << e.what() << endl;
                continue;
            }

            try {
                client_->async_get((*it), {});
                task_handler->wait_to_finished();
                if (task_handler->get_result()) {
                    cout << "download content successfully:" <<
                            (*it)->file_path() << endl;
                }
                task_handler->detach_request();
            } catch(std::domain_error &e) {
                // if error is caused by net issue,not manually(application callback)
                // increase broken counter and try it again. Once broken counter exceeds broken limit,
                // it's regarded as the "broken" one and drop it
                if (task_handler->status() != Client::TaskStatus::Canceled) {
                    if (!task_handler->is_broken()) {
                        task_handler->add_broken_counter();
                        task_handler->set_status(Client::TaskStatus::Unstart);
                    } else {
                        task_handler->set_status(Client::TaskStatus::Broken);
                    }
                }

                cerr << "download content: "<< (*it)->content_name()
                     <<  " error: " << e.what() << endl;
            }
        }
    }
}

namespace mcloud {
namespace api {

template <>
void SyncThread<UploadTask>::run() {
    std::unique_lock<std::mutex> lock(mutex_);

    while(!stop_) {
        con_.wait_for(lock, std::chrono::seconds(sync_timeout_));

        if (!stop_ && !pause_) {
                //check if there is unstared task item in task queue
                typename TaskQueue<std::shared_ptr<UploadTask>>::const_iterator it = std::find_if(
                            task_queue_.begin(),
                            task_queue_.end(),
                            [](std::shared_ptr<UploadTask> task){
                    return (task->status() == Client::TaskStatus::Unstart);
                });

                if (it == task_queue_.end()) {
                    continue;
                }

                std::shared_ptr<TaskHandler> task_handler = (*it)->task_handler();
                http::Header headers;
                task_handler->on_prepare()((void *)&headers);

                std::ifstream ifs((*it)->file_path());

                try {
                    client_->async_post((*it), headers, ifs, (*it)->file_size());
                    task_handler->wait_to_finished();
                    if (task_handler->get_result()) {
                        cout << "upload content successfully:" << (*it)->file_path() << endl;
                    }
                    task_handler->detach_request();

                    ifs.close();
                } catch(std::domain_error &e) {
                    // if error is caused by net issue,non-manually(application callback)
                    // increase broken counter and try it again. Once broken counter exceeds broken limit,
                    // it's regarded as the "broken" one and drop it
                    if (task_handler->status() != Client::TaskStatus::Canceled) {
                        if (!task_handler->is_broken()) {
                            task_handler->add_broken_counter();
                            task_handler->set_status(Client::TaskStatus::Unstart);
                        } else {
                            task_handler->set_status(Client::TaskStatus::Broken);
                        }
                    }

                    cerr << "upload content: "<< (*it)->content_name()
                         <<  " error: " << e.what() << endl;
                }
            }
        }
    }
}
}

template class SyncThread<DownloadTask>;
template class SyncThread<UploadTask>;
