/*
 * Copyright © 2016 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Gary Wang <gary.wang@canonical.com>
 */

#include <boost/filesystem.hpp>
#include <iostream>

#include <mcloud/api/uploadtask.h>

#include <tinyxml2.h>

#include "taskhandler.h"

namespace http = core::net::http;
namespace fs = boost::filesystem;

using namespace mcloud::api;
using namespace std;

class UploadTask::Priv{
public:
    Priv(const tinyxml2::XMLElement *root,
         const std::string &file_path,
         const std::string &task_id,
         const std::string &redirection_url)
        : file_path_(file_path),
          file_size_(0),
          progress_handler_(nullptr),
          status_handler_(nullptr),
          task_handler_(std::make_shared<TaskHandler>()){
        if (fs::exists(file_path_)) {
            file_size_ = fs::file_size(file_path_);
        } else {
            cerr << file_path << " doesn't exists " << endl;
            task_handler_->set_status(Client::TaskStatus::Broken);
        }

        content_id_ = root->FirstChildElement("contentID")->GetText();
        content_name_ = root->FirstChildElement("contentName")->GetText();
        is_need_upload_ = !strcmp(root->FirstChildElement("isNeedUpload")->GetText(), "1");

#ifndef NDEBUG
        cout <<"is_need_upload_:    " << is_need_upload_ << endl;
#endif

        //If file exists in cloud, then set status to complete to avoid duplicate upload
        if (!is_need_upload_) {
            task_handler_->set_status(Client::TaskStatus::Complete);
        }

        upload_task_id_ = task_id;
        task_url_ = redirection_url;

        task_handler_->on_prepare() = [this](void* params) {
            //change status
            task_handler_->set_status(Client::TaskStatus::Running);

            //populate header info
            http::Header *header = reinterpret_cast<http::Header *>(params);

            std::string content_type;
            std::string suffix = fs::path(file_path_).extension().string();
            if (suffix == "gif" || suffix == "jpg" || suffix == "jpeg" || suffix == "png") {
                content_type = "image/"+ suffix +"; name=" + content_name_;
            } else if (suffix == "avi" || suffix == "mp4" || suffix == "mkv") {
                content_type = "video/"+ suffix +"; name=" + content_name_;
            } else if (suffix == "wav" || suffix == "ogg") {
                content_type = "audio/"+ suffix +"; name=" + content_name_;
            } else if (suffix == "mp3") {
                content_type = "audio/mpeg; name=" + content_name_;
            } else if (suffix == "ppt") {
                content_type = "application/vnd.ms-powerpoint; name=" + content_name_;
            } else if (suffix == "pptx") {
                content_type = "application/vnd.openxmlformats-officedocument.presentationml.presentation; name="
                                + content_name_;
            } else if (suffix == "doc") {
                content_type = "application/msword; name=" + content_name_;
            } else if (suffix == "docx") {
                content_type =
                    "application/vnd.openxmlformats-officedocument.wordprocessingml.document; name="
                        + content_name_;
            } else if (suffix == "xls") {
                content_type = "application/vnd.ms-excel; name=" + content_name_;
            } else if (suffix == "xlsx") {
                content_type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet; name="
                        + content_name_;
            } else if (suffix == "odt") {
                content_type = "application/vnd.oasis.opendocument.text; name=" + content_name_;
            } else if (suffix == "ods") {
                content_type = "application/vnd.oasis.opendocument.spreadsheet; name=" + content_name_;
            } else if (suffix == "odp") {
                content_type = "application/vnd.oasis.opendocument.presentation; name=" + content_name_;
            } else if (suffix == "txt") {
                content_type = "text/plain; name=" + content_name_;
            } else {
                content_type = "application/octet-stream; name=" + content_name_;
            }
            header->add("Content-Type", content_type);

            //mcloud proprioty header spec
            std::string size_str = std::to_string(file_size_);
            header->add("Range", "bytes=0-" + size_str);
            header->add("RangeType", "2");
            header->add("ContentSize", size_str);
            header->add("UploadtaskID", upload_task_id_);
            header->add("Content-Length", size_str);
        };

        task_handler_->on_progress() = [this](float percent) {
            if (progress_handler_) {
                progress_handler_(percent);
            }
        };

        task_handler_->on_status() = [this](Client::TaskStatus status) {
            if (status_handler_)  {
                status_handler_(status);
            }
        };

        task_handler_->on_finished() = [this]() -> bool {
            //change status
            task_handler_->reset_broken_counter();
            task_handler_->set_status(Client::TaskStatus::Complete);

            return true;
        };

        task_handler_->on_ready_read() = [this](const string& data) {
            tinyxml2::XMLDocument doc;
            tinyxml2::XMLError error = doc.Parse(data.c_str());
            const tinyxml2::XMLElement *result = doc.FirstChildElement();

            if (error != tinyxml2::XML_SUCCESS && !strcmp(result->Attribute("resultCode"), "0")) {
                cerr << "upload content interrupted " << endl;
            }
        };
    }

    const string upload_task_id() const {
        return task_url_;
    }

    const string content_id() const {
        return content_id_;
    }

    const string content_name() const {
        return content_name_;
    }

    const string file_path() const {
        return file_path_;
    }

    const string task_url() const {
        return task_url_;
    }

    size_t file_size() const {
        return file_size_;
    }

    bool is_need_upload() const {
        return is_need_upload_;
    }

    http::Header prepare() {
        //change status
        task_handler_->set_status(Client::TaskStatus::Running);

        //populate header info
        http::Header header;

        std::string content_type;
        std::string suffix = fs::path(file_path_).extension().string();
        if (suffix == "gif" || suffix == "jpg" || suffix == "jpeg" || suffix == "png") {
            content_type = "image/"+ suffix +"; name=" + content_name_;
        } else if (suffix == "avi" || suffix == "mp4" || suffix == "mkv") {
            content_type = "video/"+ suffix +"; name=" + content_name_;
        } else if (suffix == "wav" || suffix == "ogg") {
            content_type = "audio/"+ suffix +"; name=" + content_name_;
        } else if (suffix == "mp3") {
            content_type = "audio/mpeg; name=" + content_name_;
        } else if (suffix == "ppt") {
            content_type = "application/vnd.ms-powerpoint; name=" + content_name_;
        } else if (suffix == "pptx") {
            content_type = "application/vnd.openxmlformats-officedocument.presentationml.presentation; name="
                            + content_name_;
        } else if (suffix == "doc") {
            content_type = "application/msword; name=" + content_name_;
        } else if (suffix == "docx") {
            content_type =
                "application/vnd.openxmlformats-officedocument.wordprocessingml.document; name="
                    + content_name_;
        } else if (suffix == "xls") {
            content_type = "application/vnd.ms-excel; name=" + content_name_;
        } else if (suffix == "xlsx") {
            content_type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet; name="
                    + content_name_;
        } else if (suffix == "odt") {
            content_type = "application/vnd.oasis.opendocument.text; name=" + content_name_;
        } else if (suffix == "ods") {
            content_type = "application/vnd.oasis.opendocument.spreadsheet; name=" + content_name_;
        } else if (suffix == "odp") {
            content_type = "application/vnd.oasis.opendocument.presentation; name=" + content_name_;
        } else if (suffix == "txt") {
            content_type = "text/plain; name=" + content_name_;
        } else {
            content_type = "application/octet-stream; name=" + content_name_;
        }
        header.add("Content-Type", content_type);

        //mcloud proprioty header spec
        boost::uintmax_t size = file_size_;
        std::string size_str = std::to_string(size);
        header.add("Range", "bytes=0-" + size_str);
        header.add("RangeType", "2");
        header.add("ContentSize", size_str);
        header.add("UploadtaskID", upload_task_id_);
        header.add("Content-Length", size_str);

        return header;
    }

    std::string upload_task_id_;

    std::string task_url_;

    std::string file_path_;

    std::string content_id_;

    std::string content_name_;

    size_t      file_size_;

    bool        is_need_upload_;

    std::ifstream   ifs_;

    UploadTask::ProgressHandler progress_handler_;

    UploadTask::StatusHandler status_handler_;

    std::shared_ptr<TaskHandler> task_handler_;
};

UploadTask::UploadTask(const tinyxml2::XMLElement *root,
                       const std::string &file_path,
                       const std::string &task_id,
                       const std::string &redirection_url)
    : p(std::make_shared<Priv>(root, file_path, task_id, redirection_url)){
}

std::shared_ptr<TaskHandler> UploadTask::task_handler() {
    return p->task_handler_;
}

const string UploadTask::upload_task_id() const {
    return p->task_url_;
}

const string UploadTask::content_id() const {
    return p->content_id_;
}

const string UploadTask::content_name() const {
    return p->content_name_;
}

const string UploadTask::file_path() const {
    return p->file_path_;
}

const string UploadTask::task_url() const {
    return p->task_url_;
}

Client::TaskStatus UploadTask::status() const {
    return p->task_handler_->status();
}

size_t UploadTask::file_size() const {
    return p->file_size_;
}

bool UploadTask::is_need_upload() const {
    return p->is_need_upload_;
}

UploadTask::ProgressHandler &UploadTask::progress_changed() {
    return p->progress_handler_;
}

UploadTask::StatusHandler &UploadTask::status_changed() {
    return p->status_handler_;
}
