#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 Canonical, Ltd.
# 
# This library is free software; you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Gary Wang  <gary.wang@canonical.com>

import json
import os
import sys
import tornado.httpserver
import tornado.ioloop
import tornado.netutil
import tornado.web
import xml.dom.minidom
import base64
import random

ACCESS_TOKEN = 'the_access_token'

SOCK_PORT = 0

AUTHORIZATION_BEARER = 'Bearer %s' % ACCESS_TOKEN

def read_file(path):
    file = os.path.join(os.path.dirname(__file__), path)
    if os.path.isfile(file):
        with open(file, 'r') as fp:
            content = fp.read()
    else:
        raise Exception("File '%s' not found\n" % file)
    return content

def create_file(path):
    file = os.path.join(os.path.dirname(__file__), path)
    with open(file, "wb") as fp:
        size = random.randint(1024, 1024 * 1024 * 128)
        fp.truncate(size)

class ErrorHandler(tornado.web.RequestHandler):
    def write_error(self, status_code, **kwargs):
        self.write(json.dumps({'error': '%s: %d' % (kwargs["exc_info"][1], status_code)}))

class DiskInfo(ErrorHandler):
    def post(self):
        check_header(self, 'Content-Type', 'text/xml;UTF-8')
        check_header(self, 'Content-Length')

        file = 'diskinfo/diskinfo.xml'
        self.write(read_file(file))
        self.finish()

class ResourceList(ErrorHandler):
    def post(self):
        check_header(self, 'Content-Type', 'text/xml;UTF-8')
        check_header(self, 'Content-Length')

        dom = xml.dom.minidom.parseString(self.request.body)
        disk = dom.getElementsByTagName("getDisk")[0]
        catalogId = disk.getElementsByTagName("catalogID")[0].firstChild.data
        
        file = 'resources/%s.xml' % catalogId
        self.write(read_file(file))
        self.finish()

class DeleteResource(ErrorHandler):
    def post(self):
        check_header(self, 'Content-Type', 'text/xml;UTF-8')
        check_header(self, 'Content-Length')

        dom = xml.dom.minidom.parseString(self.request.body)
        disk = dom.getElementsByTagName('delCatalogContent')[0]
        catalogIdList = disk.getElementsByTagName('catalogIDs')
        contentIdList = disk.getElementsByTagName('contentIDs')

        if int(catalogIdList[0].attributes['length'].value) > 0:
            category = "folder"
            folderId = catalogIdList[0].getElementsByTagName('ID')[0].firstChild.data
        if int(contentIdList[0].attributes['length'].value) > 0:
            category = "content"
            folderId = contentIdList[0].getElementsByTagName('ID')[0].firstChild.data

        file = 'resources/%s_%s.xml' % (category, folderId)
        self.write(read_file(file))
        self.finish()

class Outlinks(ErrorHandler):
    def post(self):
        check_header(self, 'Content-Type', 'text/xml;UTF-8')
        check_header(self, 'Content-Length')

        dom = xml.dom.minidom.parseString(self.request.body)
        outLink = dom.getElementsByTagName('getOutLink')[0]
        outLinkReq  = outLink.getElementsByTagName('getOutLinkReq')[0]
        catalogIdList = outLinkReq.getElementsByTagName('caIDLst')
        contentIdList = outLinkReq.getElementsByTagName('coIDLst')

        if int(catalogIdList[0].attributes['length'].value) > 0:
            category = "folder"
            folderId = catalogIdList[0].getElementsByTagName('item')[0].firstChild.data
        if int(contentIdList[0].attributes['length'].value) > 0:
            category = "content"
            folderId = contentIdList[0].getElementsByTagName('item')[0].firstChild.data

        file = 'resources/outlink_%s_%s.xml' % (category, folderId)
        self.write(read_file(file))
        self.finish()

class UploadAndDownload(ErrorHandler):
    def post(self):
        check_header(self, 'Content-Type', 'text/xml;UTF-8')
        check_header(self, 'Content-Length')

        #contentName = ""
        dom = xml.dom.minidom.parseString(self.request.body)
        uploadRequests = dom.getElementsByTagName('pcUploadFileRequest')
        if uploadRequests:
            uploadContentList  = uploadRequests[0].getElementsByTagName('uploadContentList')[0]
            uploadContentInfo = uploadContentList.getElementsByTagName('uploadContentInfo')[0]
            contentName = uploadContentInfo.getElementsByTagName('contentName')[0].firstChild.data

        downloadRequest = dom.getElementsByTagName('downloadRequest')
        if downloadRequest:
            contentName = downloadRequest[0].getElementsByTagName('contentID')[0].firstChild.data

        file = 'resources/%s.xml' % (contentName)
        content = read_file(file)
        content = content.replace('[port]', str(SOCK_PORT))
        self.write(content)
        self.finish()

class DownloadFile(ErrorHandler):
    def get(self):
        data = self.request.arguments['filename'][0]
        data = data.replace(' ', '+') 
        contentName = str(data).decode('base64')

        file = 'files/%s' % (contentName)
        create_file(file)
        self.write(read_file(file))
        self.finish()

class UploadFile(ErrorHandler):
    def post(self):
        file = 'resources/upload_response.xml'
        self.write(read_file(file))
        self.finish()

def check_header(self, name, expected=None):
    actual = self.request.headers.get(name, '')
    if expected == None and int(actual) < 0:
        raise Exception("Header '%s' == '%s' < 0" % (name, actual))
    elif expected != None and actual != expected:
        raise Exception("Header '%s' == '%s' != '%s'" % (name, actual, expected))

def new_app():
    application = tornado.web.Application([
        (r"/richlifeApp/devapp/IUser", DiskInfo),
        (r"/richlifeApp/devapp/ICatalog", ResourceList),
        (r"/richlifeApp/devapp/IContent", DeleteResource),
        (r"/richlifeApp/devapp/IOutLink", Outlinks),
        (r"/richlifeApp/devapp/IUploadAndDownload", UploadAndDownload),
        (r"/storageWeb/servlet/pcDownloadFile", DownloadFile),
        (r"/storageWeb/servlet/uploadFileServlet", UploadFile)
    ])
    sockets = tornado.netutil.bind_sockets(0, '127.0.0.1')
    server = tornado.httpserver.HTTPServer(application)
    server.add_sockets(sockets)

    global SOCK_PORT
    SOCK_PORT = sockets[0].getsockname()[1]
    sys.stdout.write('%d\n' % SOCK_PORT)
    sys.stdout.flush()

    return application

if __name__ == "__main__":
    application = new_app()
    tornado.ioloop.IOLoop.instance().start()
