#include <core/posix/exec.h>

#include <mcloud/api/client.h>
#include <mcloud/api/diskinfo.h>
#include <mcloud/api/uploadtask.h>
#include <mcloud/api/downloadtask.h>
#include <mcloud/api/outlink.h>
#include <gtest/gtest.h>

#include <iostream>
#include <random>

using namespace mcloud::api;
using namespace core::posix;
using namespace std;

static const std::string TMPDIR = "/tmp";

static const char * Status[] = { "unstart", "running", "canceled", "paused", "broken", "complete"};

namespace {
    std::string tmpdir() {
        const char *v = getenv("TMPDIR");
        return v == NULL ? TMPDIR : std::string(v);
    }

    std::string status_to_string(Client::TaskStatus status) {
        return Status[(int)status];
    }

    void create_file(std::string file_path, size_t size) {
        std::ofstream ofs(file_path, std::ios::binary | std::ios::out);
        ofs.seekp(size);
        ofs.write("", 1);
        ofs.close();
    }

    class McloudAPI : public ::testing::Test {
    protected:
        virtual void SetUp() override {
            fake_mcloud_server_ = exec(FAKE_MCLOUD_SERVER, { }, { },
                    StandardStream::stdout);

            ASSERT_GT(fake_mcloud_server_.pid(), 0);
            string port;
            fake_mcloud_server_.cout() >> port;

            string apiroot = "http://127.0.0.1:" + port;
            setenv("MCLOUD_LOCAL_SERVER_URL", apiroot.c_str(), true);

            string download_folder = tmpdir();
            setenv("MCLOUD_DOWNLOAD_FOLDER", download_folder.c_str(), true);
        }

        virtual void TearDown() override {
        }

        ChildProcess fake_mcloud_server_ = ChildProcess::invalid();
    };

TEST_F(McloudAPI, disk_info)
{
    const int time_out = 10;
    Client c(time_out);

    auto disk_info = c.disk_info();
    EXPECT_EQ(14600, disk_info->free_disk_size());
    EXPECT_EQ(20480, disk_info->disk_size());
}

TEST_F(McloudAPI, share_folder_id)
{
    const int time_out = 10;
    Client c(time_out);

    auto share_folder_id = c.cloud_sync_folder_id();
    EXPECT_EQ("1011YfnDD09x00019700101000000001", share_folder_id);
}

TEST_F(McloudAPI, cloud_content_list)
{
    const int time_out = 10;
    Client c(time_out);

    const int start_index = 0;
    const int count = 200;
    auto share_folder_id = c.cloud_sync_folder_id();
    auto content_list = c.cloud_content_list(start_index, 
            count, Client::ContentType::all, share_folder_id);

    EXPECT_EQ("1011YfnDD09x00019700101000000001", share_folder_id);
    EXPECT_EQ(46, content_list.size());
}

TEST_F(McloudAPI, create_folder_extranet_link)
{
    const int time_out = 10;
    Client c(time_out);

    Client::Stringlist folderlist;
    string folder_id = "1011YfnDD09x03720160403231108kn4";
    folderlist.push_back(folder_id);
    auto outlink_list = c.create_folder_extranet_link(folderlist);

    EXPECT_EQ(1, outlink_list.size());
    EXPECT_EQ("http://caiyun.feixin.10086.cn/dl/105CpSe8FosMu", outlink_list[0]->link());
}

TEST_F(McloudAPI, create_content_extranet_link)
{
    const int time_out = 10;
    Client c(time_out);

    Client::Stringlist contentlist;
    string content_id = "1011YfnDD09x03620151105110237b7o";
    contentlist.push_back(content_id);
    auto outlink_list = c.create_content_extranet_link(contentlist);

    EXPECT_EQ(1, outlink_list.size());
    EXPECT_EQ("http://caiyun.feixin.10086.cn/dl/105CqR96ak9mk", outlink_list[0]->link());
}

TEST_F(McloudAPI, delete_folder)
{
    const int time_out = 10;
    Client c(time_out);

    Client::Stringlist folderlist;
    string folder_id = "1011YfnDD09x03620160403222045id6";
    folderlist.push_back(folder_id);
    auto is_removed = c.delete_folders(folderlist);

    EXPECT_EQ(true, is_removed);
}

TEST_F(McloudAPI, delete_content)
{
    const int time_out = 10;
    Client c(time_out);

    Client::Stringlist contentlist;
    string content_id = "1011YfnDD09x03920151110171910rox";
    contentlist.push_back(content_id);
    auto is_removed = c.delete_contents(contentlist);

    EXPECT_EQ(true, is_removed);
}

TEST_F(McloudAPI, no_need_upload_on_cloud)
{
    const std::size_t size = 1024;
    static const std::string file_path = tmpdir() + "/file_exists_on_cloud";
    create_file(file_path, size);

    const int time_out = 10;
    Client c(time_out);
    auto is_exist = c.exist_on_cloud(file_path);

    EXPECT_EQ(true, is_exist);
}

TEST_F(McloudAPI, need_upload_on_cloud)
{
    const std::size_t size = 1024;
    static const std::string file_path = tmpdir() + "/file_not_exists_on_cloud";
    create_file(file_path, size);

    const int time_out = 10;
    Client c(time_out);
    auto is_exist = c.exist_on_cloud(file_path);

    EXPECT_EQ(false, is_exist);
}

TEST_F(McloudAPI, file_not_exists_on_local)
{
    static constexpr const char *file_name = "file_not_exists_on_local";
    std::vector<std::string> file_path_list{file_name};

    const int time_out = 10;
    Client c(time_out);
    c.add_upload_tasks(file_path_list);

    TaskQueue<Client::UploadTaskPtr> upload_list = c.upload_queue();
    EXPECT_EQ(1, upload_list.size());
    EXPECT_EQ(file_name, upload_list[0]->content_name());
    EXPECT_EQ(Client::TaskStatus::Broken, upload_list[0]->status());
}

TEST_F(McloudAPI, invalid_content_id)
{
    static constexpr const char *invalid_content_id = "1011YfnD00000000000000000000000000";
    std::vector<std::string> content_id_list{invalid_content_id};

    const int time_out = 10;
    Client c(time_out);
    c.add_download_tasks(content_id_list);

    TaskQueue<Client::DownloadTaskPtr> download_list = c.download_queue();
    EXPECT_EQ(0, download_list.size());
}

TEST_F(McloudAPI, sync_download_files)
{
    static constexpr const char *mp3_content_id = "1811asktx23a054201604201417321zo";
    static constexpr const char *mp4_content_id = "1811asktx23a05620160425125424fle";
    static constexpr const char *mp4_content_id_2 = "1811asktx23a054201604201417331zp";
    static constexpr const char *gif_content_id = "1811asktx23a0572016042014173129t";
    static constexpr const char *pdf_content_id = "1811asktx23a05620160420134213zx4";
    static constexpr const char *png_content_id = "1811asktx23a058201604201519046ec";

    std::vector<std::string> content_id_list;
    content_id_list.push_back(mp3_content_id);
    content_id_list.push_back(mp4_content_id);
    content_id_list.push_back(mp4_content_id_2);
    content_id_list.push_back(gif_content_id);
    content_id_list.push_back(pdf_content_id);
    content_id_list.push_back(png_content_id);

    const int time_out = 10;
    Client c(time_out);
    Client::DownloadList added_download_tasks = c.add_download_tasks(content_id_list);

    //Client::DownloadList download_list = c.download_queue();
    EXPECT_EQ(6, added_download_tasks.size());

    for (auto & task : added_download_tasks) 
    {
        task->progress_changed() = [&task](float percent) {
            cout<< "downloading: "<< task->content_name() << "  "<< percent << endl;
        };

        task->status_changed() = [&task](Client::TaskStatus status) {
            cout<< "status : "<< status_to_string(status) << "  "<< task->content_name() << endl;
        };
    }

    std::this_thread::sleep_for(std::chrono::milliseconds(2000 * added_download_tasks.size()));

    EXPECT_EQ(Client::TaskStatus::Complete, added_download_tasks[0]->status());
    EXPECT_EQ(Client::TaskStatus::Complete, added_download_tasks[1]->status());
    EXPECT_EQ(Client::TaskStatus::Complete, added_download_tasks[2]->status());
    EXPECT_EQ(Client::TaskStatus::Complete, added_download_tasks[3]->status());
    EXPECT_EQ(Client::TaskStatus::Complete, added_download_tasks[4]->status());
    EXPECT_EQ(Client::TaskStatus::Complete, added_download_tasks[5]->status());
}

TEST_F(McloudAPI, sync_upload_files)
{
    static constexpr const char *mp3_upload_content = "song.mp3";
    static constexpr const char *mp4_upload_content = "movie.mp4";
    static constexpr const char *gif_upload_content = "funny.gif";
    static constexpr const char *pdf_upload_content = "ubuntu.pdf";
    static constexpr const char *png_upload_content = "screenshot.png";

    std::vector<std::string> local_file_list;
    local_file_list.push_back(tmpdir() + "/" + mp3_upload_content);
    local_file_list.push_back(tmpdir() + "/" + mp4_upload_content);
    local_file_list.push_back(tmpdir() + "/" + gif_upload_content);
    local_file_list.push_back(tmpdir() + "/" + pdf_upload_content);
    local_file_list.push_back(tmpdir() + "/" + png_upload_content);

    std::default_random_engine engine;
    std::uniform_int_distribution<int> dis(1024, 1024 * 1024 * 50);
    for (auto file_path : local_file_list) {
        size_t file_size = dis(engine);
        create_file(file_path, file_size);
    }

    const int time_out = 10;
    Client c(time_out);
    Client::UploadList added_upload_tasks = c.add_upload_tasks(local_file_list);

    //Client::UploadList upload_list = c.upload_queue();
    EXPECT_EQ(5, added_upload_tasks.size());

    for (auto & task : added_upload_tasks) 
    {
        task->progress_changed() = [&task](float percent) {
            cout << "uploading: " << task->content_name() << "  "<< percent << endl;
        };

        task->status_changed() = [&task](Client::TaskStatus status) {
            cout<< "status : "<< status_to_string(status) << "  "<< task->content_name() << endl;
        };
    }
    
    std::this_thread::sleep_for(std::chrono::milliseconds(2000 * added_upload_tasks.size()));

    EXPECT_EQ(Client::TaskStatus::Complete, added_upload_tasks[0]->status());
    EXPECT_EQ(Client::TaskStatus::Complete, added_upload_tasks[1]->status());
    EXPECT_EQ(Client::TaskStatus::Complete, added_upload_tasks[2]->status());
    EXPECT_EQ(Client::TaskStatus::Complete, added_upload_tasks[3]->status());
    EXPECT_EQ(Client::TaskStatus::Complete, added_upload_tasks[4]->status());
}

}

